import re


def to_cmd_name(name: str) -> str:
    return re.sub(r'[\W]+', '-', name).strip('-').lower()


def to_variable_name(string):
    # Remove any non-alphanumeric characters
    string = re.sub(r'[^0-9a-zA-Z]+', ' ', string).strip()
    # Convert to lowercase
    string = string.lower()
    # Replace spaces with underscores
    string = string.replace(' ', '_')
    # Remove leading digits
    string = re.sub(r'^[0-9]+', '', string)
    return string



from zrb.builtin._group import project_group
from zrb import Task, runner

deploy_project = Task(
    name='deploy',
    group=project_group,
    upstreams=[],
    description='Deploy project'
)
runner.register(deploy_project)



from zrb_init import concat, register_trainer, fibo

print('Result', concat.create_main_loop()(
    'angry', 'bulbasaur', separator='-'
))

print('Result', register_trainer.create_main_loop()(
    name='John',
    password='secret',
    age=10,
    starter_pokemon='bulbasaur'
))

print('Result', fibo.create_main_loop()(
    n=7
))



from zrb import (
    runner, CmdTask, StrInput, HTTPChecker
)

build_task = CmdTask(
    name='build',
    description='Build Zrb',
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Build zrb distribution"',
        'rm -Rf ${ZRB_PROJECT_DIR}/dist',
        'git add . -A',
        'flit build',
    ],
)
runner.register(build_task)

publish_task = CmdTask(
    name='publish',
    description='Publish zrb to pypi',
    upstreams=[build_task],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Publish zrb to pypi"',
        'flit publish --repository pypi',
    ]
)
runner.register(publish_task)

publish_test_task = CmdTask(
    name='publish-test',
    description='Publish zrb to testpypi',
    upstreams=[build_task],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Publish zrb to testpypi"',
        'flit publish --repository testpypi',
    ]
)
runner.register(publish_test_task)

install_symlink_task = CmdTask(
    name='install-symlink',
    description='Install Zrb as symlink',
    upstreams=[build_task],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Install zrb"',
        'flit install --symlink',
    ]
)
runner.register(install_symlink_task)

test_task = CmdTask(
    name='test',
    description='Run zrb test',
    inputs=[
        StrInput(
            name='test',
            shortcut='t',
            prompt='Test (i.e., test/file.py::test_name)',
            default=''
        )
    ],
    upstreams=[install_symlink_task],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Perform test"',
        ' '.join([
            'pytest --cov=zrb --cov-report html',
            '--cov-report term --cov-report term-missing',
            '{{input.test}}'
        ])
    ],
    retry=0,
    checking_interval=1
)
runner.register(test_task)

serve_test_task = CmdTask(
    name='serve-test',
    description='Serve zrb test result',
    inputs=[
        StrInput(
            name='port',
            shortcut='p',
            prompt='Serve coverage on port',
            default='9000'
        )
    ],
    upstreams=[test_task],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Serve coverage report"',
        ' '.join([
            'python -m http.server {{input.port}}',
            '--directory ${ZRB_PROJECT_DIR}/htmlcov',
        ])
    ],
    checkers=[
        HTTPChecker(port='{{input.port}}')
    ],
    retry=0,
    checking_interval=0.3
)
runner.register(serve_test_task)

prepare_playground_cmds = [
    'set -e',
    'cd ${ZRB_PROJECT_DIR}',
    'if [ ! -d playground ]',
    'then',
    '  echo "🤖 Create playground"',
    '  cp -R playground-template playground',
    'fi',
    'cd ${ZRB_PROJECT_DIR}/playground',
    'echo "🤖 Run zrb in the playground"',
    'zrb',
    'echo "🤖 Invoke the following command:"',
    'echo "      cd ${ZRB_PROJECT_DIR}/playground"',
    'echo "🤖 And start hacking around. Good luck :)"',
]


prepare_playground_task = CmdTask(
    name='prepare-playground',
    upstreams=[install_symlink_task],
    cmd=prepare_playground_cmds
)
runner.register(prepare_playground_task)


reset_playground_task = CmdTask(
    name='reset-playground',
    description='Reset Zrb playground',
    upstreams=[install_symlink_task],
    cmd=[
        'set -e',
        'cd ${ZRB_PROJECT_DIR}',
        'echo "🤖 Remove playground"',
        'rm -Rf playground',
    ] + prepare_playground_cmds
)
runner.register(reset_playground_task)



import libcst as cst


def add_upstream_to_task(
    code: str, task_name: str, upstream_task_name: str
) -> str:
    """
    Transformer to add an upstream task to a Task object.

    Args:
        code (str): The code to modify.
        task_name (str): Name of the task to modify.
        upstream_task_name (str): Name of the upstream task to add.

    Returns:
        A modified version of the input code with the upstream task added.
    """

    class TaskTransformer(cst.CSTTransformer):
        def __init__(self, task_name: str, upstream_task_name: str):
            super().__init__()
            self.task_name: str = task_name
            self.upstream_task_name: str = upstream_task_name
            self._on_task_call = False
            self._upstream_added = False

        def _is_task(self, node: cst.Call) -> bool:
            if (
                not isinstance(node.func, cst.Name)
                or node.func.value != "Task"
            ):
                return False
            # Look for name keyword
            for name_arg in node.args:
                if (
                    isinstance(name_arg, cst.Arg)
                    and name_arg.keyword.value == "name"
                    and name_arg.value.evaluated_value == self.task_name
                ):
                    return True
            return False

        def visit_Call(self, node: cst.Call):
            # Visit all function call (or class initiation)
            # If it is the task we seek, flag it
            if self._is_task(node):
                self._on_task_call = True

        def leave_Call(
            self, original_node: cst.Call, updated_node: cst.Call
        ) -> cst.Call:
            # Remove flags.
            # If upstreams argument not found (see: self.leave_Arg), add it
            if not self._is_task(updated_node):
                return updated_node
            self._on_task_call = False
            if self._upstream_added:
                return updated_node
            new_args = [arg for arg in updated_node.args]
            new_args.append(cst.Arg(
                keyword=cst.parse_expression('upstreams'),
                value=cst.List([
                    cst.Element(cst.parse_expression(self.upstream_task_name))
                ])
            ))
            self._upstream_added = True
            return updated_node.with_changes(
                args=new_args
            )

        def leave_Arg(self, orginnal_node: cst.Arg, updated_node: cst.Arg):
            # Once the flag was set (see: self.visit_Call), look for
            # `upstreams` argument and modify the value
            if (
                not self._on_task_call
                or updated_node.keyword.value != 'upstreams'
            ):
                return updated_node
            new_elements = [
                old_element for old_element in updated_node.value.elements
            ]
            new_elements.append(cst.Element(
                value=cst.parse_expression(self.upstream_task_name)
            ))
            self._upstream_added = True
            return updated_node.with_changes(
                value=cst.List(new_elements)
            )

    transformer: TaskTransformer = TaskTransformer(
        task_name, upstream_task_name
    )
    original_tree: cst.Module = cst.parse_module(code)
    modified_tree: cst.Module = original_tree.visit(transformer)
    return modified_tree.code



from typing import List
from ...helper.middlewares.replacement import (
    Replacement, ReplacementMiddleware
)
from ...task_input.str_input import StrInput
from ...task_input.int_input import IntInput
from ...helper import util
from ...helper.accessories.name import get_random_name
from ...helper.codemod.add_assert_resource import add_assert_resource
from ...helper.codemod.add_import_module import add_import_module
from ...helper.middlewares.replacement import (
    add_pascal_key, add_snake_key, add_camel_key,
    add_kebab_key, add_human_readable_key
)
import os


project_dir_input = StrInput(
    name='project-dir',
    shortcut='d',
    prompt='Project directory',
    default='.'
)

project_name_input = StrInput(
    name='project-name',
    shortcut='n',
    prompt='Project name (can be empty)',
    default=''
)

app_name_input = StrInput(
    name='app-name',
    shortcut='a',
    prompt='App name',
    default=get_random_name()
)

task_name_input = StrInput(
    name='task-name',
    shortcut='t',
    prompt='Task name',
    default=f'run-{get_random_name()}'
)
http_port_input = IntInput(
    name='http-port',
    shortcut='p',
    prompt='Port',
    default=8080
)

new_task_scaffold_lock = os.path.sep.join([
    '{{ os.path.join(input.project_dir, "_automate") }}',
    '{{ util.to_snake_case(input.task_name) }}.py'
])

new_app_scaffold_lock = os.path.sep.join([
    '{{ os.path.join(input.project_dir, "_automate") }}',
    '{{ util.to_snake_case(input.app_name) }}.py'
])


def validate_project_dir(project_dir: str):
    if not os.path.isfile(os.path.join(project_dir, 'zrb_init.py')):
        raise Exception(f'Not a project: {project_dir}')


def get_automation_file_name(
    zrb_project_dir: str, automation_name: str
) -> str:
    snake_automation_name = util.to_snake_case(automation_name)
    return os.path.join(
        zrb_project_dir,
        '_automate',
        f'{snake_automation_name}.py'
    )


def validate_automation_name(project_dir: str, automation_name: str):
    automation_file = get_automation_file_name(project_dir, automation_name)
    if os.path.isfile(automation_file):
        raise Exception(f'File already exists: {automation_file}')


def get_automation_module_name(automation_name: str):
    snake_automation_name = util.to_snake_case(automation_name)
    return '.'.join(['_automate', snake_automation_name])


def register_module(project_dir: str, module_name: str):
    zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
    with open(zrb_init_path, 'r') as f:
        code = f.read()
        import_alias = module_name.split('.')[-1]
        code = add_import_module(
            code=code,
            module_path=module_name,
            alias=import_alias
        )
        code = add_assert_resource(code, import_alias)
    with open(zrb_init_path, 'w') as f:
        f.write(code)
    return True


def get_default_task_replacements() -> Replacement:
    return {
        'taskName': '{{input.task_name}}',
        'appName': '{{input.app_name}}',
        'httpPort': '{{util.coalesce(input.http_port, "3000")}}',
    }


def get_default_task_replacement_middlewares() -> List[ReplacementMiddleware]:
    return [
        # task
        add_pascal_key('PascalTaskName', 'taskName'),
        add_camel_key('camelTaskName', 'taskName'),
        add_snake_key('snake_task_name', 'taskName'),
        add_kebab_key('kebab-task-name', 'taskName'),
        add_human_readable_key('human readable task name', 'taskName'),
        # app
        add_pascal_key('PascalAppName', 'appName'),
        add_camel_key('camelAppName', 'appName'),
        add_snake_key('snake_app_name', 'appName'),
        add_kebab_key('kebab-app-name', 'appName'),
        add_human_readable_key('human readable app name', 'appName'),
    ]




from zrb.builtin._group import project_group
from zrb import Task, runner

start_project_containers = Task(
    name='start-containers',
    group=project_group,
    upstreams=[],
    description='Start as containers'
)
runner.register(start_project_containers)



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common import (
    project_dir_input, task_name_input, validate_project_dir,
    get_automation_module_name, validate_automation_name, register_module,
    get_default_task_replacements, get_default_task_replacement_middlewares,
    new_task_scaffold_lock
)

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='task-validate-create',
    inputs=[project_dir_input, task_name_input],
)
def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    task_name = kwargs.get(project_dir, 'task_name')
    validate_automation_name(project_dir, task_name)


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[project_dir_input, task_name_input],
    upstreams=[validate],
    replacements=get_default_task_replacements(),
    replacement_middlewares=get_default_task_replacement_middlewares(),
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    scaffold_locks=[new_task_scaffold_lock]
)


@python_task(
    name='cmd-task',
    group=project_add_group,
    inputs=[project_dir_input, task_name_input],
    upstreams=[copy_resource],
    runner=runner
)
def add_cmd_task(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    task_name = kwargs.get('task_name')
    task.print_out(f'Register cmd task: {task_name}')
    module_name = get_automation_module_name(task_name)
    register_module(project_dir, module_name)



from zrb.task.task import Task
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_input.str_input import StrInput
import os


def test_task_with_no_runner():
    task = Task(
        name='simple-no-error',
        retry=0
    )
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result


def test_task_with_predefined_runner():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        run=_run,
        retry=0
    )
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result == 'hello'


def test_task_with_input():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        favorite_drink = kwargs['favorite_drink']
        return f'hello {name}, your favorite drink is {favorite_drink}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name'),
            StrInput(name='favorite-drink')
        ],
        run=_run,
        retry=0
    )
    main_loop = task.create_main_loop()
    result = main_loop(name='Dumbledore', favorite_drink='Elixir')
    assert result == 'hello Dumbledore, your favorite drink is Elixir'


def test_task_with_templated_env():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('__ZRB_TEST_ENV_2')
    task = Task(
        name='templated-env',
        envs=[
            Env(name='__ZRB_TEST_ENV_1', default='Elixir'),
            Env(
                name='__ZRB_TEST_ENV_2',
                default='{{env.__ZRB_TEST_ENV_1}} of immortality'
            )
        ],
        run=_run,
        retry=0
    )
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result == 'Elixir of immortality'


def test_task_with_env_file():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('__COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file',
        env_files=[EnvFile(env_file=env_file)],
        run=_run,
        retry=0
    )
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result == 'blue'


def test_task_with_env_file_and_prefix():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('__COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file-prefixed',
        env_files=[EnvFile(env_file=env_file, prefix='__ZRB_TEST')],
        run=_run,
        retry=0
    )
    main_loop = task.create_main_loop()
    os.environ['__ZRB_TEST___COLOR'] = 'red'
    result = main_loop()
    assert result == 'red'



from ....helper.file.copy_tree import copy_tree
from ....helper.codemod.add_assert_resource import add_assert_resource
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.add_upstream_to_task import add_upstream_to_task

import os

current_dir = os.path.dirname(__file__)


def add_project_automation(project_dir: str):
    if os.path.exists(os.path.join(project_dir, '_automate', '_project')):
        return
    copy_tree(
        src=os.path.join(current_dir, 'template'),
        dst=project_dir
    )
    project_task_module_path = '_automate._project'
    zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
    with open(zrb_init_path, 'r') as f:
        code = f.read()
        import_alias = project_task_module_path.split('.')[-1]
        code = add_import_module(
            code=code,
            module_path=project_task_module_path,
            alias=import_alias
        )
        code = add_assert_resource(code, import_alias)
    with open(zrb_init_path, 'w') as f:
        f.write(code)


def register_project_upstream(
    project_dir: str,
    project_automation_file: str,
    project_automation_task_name: str,
    upstream_task_file: str,
    upstream_task_var: str
):
    '''
    Adding upstream_task_var as project_task's
    '''
    project_automation_dir = os.path.join(
        project_dir, '_automate', '_project'
    )
    project_automation_path = os.path.join(
        project_automation_dir, f'{project_automation_file}'
    )
    upstream_task_rel_file_path = os.path.relpath(
        upstream_task_file, project_automation_path
    )
    # normalize `..` parts
    upstream_module_parts = [
        part if part != '..' else ''
        for part in upstream_task_rel_file_path.split(os.path.sep)
    ]
    # remove .py extenstion
    upstream_module_parts[-1] = os.path.splitext(upstream_module_parts[-1])[0]
    upstream_module_path = '.'.join(upstream_module_parts)
    with open(project_automation_path, 'r') as f:
        code = f.read()
        code = add_import_module(
            code=code,
            module_path=upstream_module_path,
            resource=upstream_task_var
        )
        code = add_upstream_to_task(
            code, project_automation_task_name, upstream_task_var
        )
    with open(project_automation_path, 'w') as f:
        f.write(code)



from typing import (
    Any, Callable, Iterable, Mapping, Optional, Tuple, TypeVar
)
from typeguard import typechecked
from .base_model import TaskModel
from ..task_input.base_input import BaseInput
from ..task_env.env import Env
from ..task_env.env_file import EnvFile
from ..task_group.group import Group
from ..helper.list.append_unique import append_unique
from ..helper.string.conversion import to_variable_name
from ..helper.advertisement import get_advertisement
from ..advertisement import advertisements


import asyncio
import copy

TTask = TypeVar('TTask', bound='BaseTask')


@typechecked
class BaseTask(TaskModel):
    '''
    Base class for all tasks.
    Every task definition should be extended from this class.
    '''

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[BaseInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[TTask] = [],
        checkers: Iterable[TTask] = [],
        checking_interval: float = 0.1,
        retry: int = 2,
        retry_interval: float = 1,
        run: Optional[Callable[..., Any]] = None
    ):
        TaskModel.__init__(
            self,
            name=name,
            group=group,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            retry=retry
        )
        retry_interval = self.ensure_non_negative(
            retry_interval, 'Find negative retry_interval'
        )
        checking_interval = self.ensure_non_negative(
            checking_interval, 'Find negative checking_interval'
        )
        self.inputs = inputs
        self.description = description
        self.retry_interval = retry_interval
        self.upstreams = upstreams
        self.checkers = checkers
        self.checking_interval = checking_interval
        self._is_checked: bool = False
        self._is_executed: bool = False
        self._runner: Optional[Callable[..., Any]] = run

    async def run(self, *args: Any, **kwargs: Any) -> Any:
        '''
        Do task execution
        Please override this method.
        '''
        if self._runner is not None:
            return self._runner(*args, **kwargs)
        return True

    async def check(self) -> bool:
        '''
        Return true when task is considered completed.
        By default, this will wait the task execution to be completed.
        You can override this method.
        '''
        return self.is_done()

    def get_all_inputs(self) -> Iterable[BaseInput]:
        ''''
        Getting all inputs of this task and all its upstream, non-duplicated.
        '''
        inputs: Iterable[BaseInput] = []
        for upstream in self.upstreams:
            upstream_inputs = upstream.get_all_inputs()
            append_unique(inputs, *upstream_inputs)
        append_unique(inputs, *self.inputs)
        return inputs

    def get_description(self) -> str:
        if self.description != '':
            return self.description
        return self.name

    def create_main_loop(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., Any]:
        def main_loop(*args: Any, **kwargs: Any) -> Any:
            '''
            Task main loop.
            '''
            self.start_timer()
            self_cp, args, kwargs = self._get_main_loop_variables(
                env_prefix=env_prefix, args=args, kwargs=kwargs
            )

            async def run_and_check_all_async() -> Any:
                # initiate args
                processes = [
                    asyncio.create_task(self_cp._loop_check(True)),
                    asyncio.create_task(self_cp._run_all(*args, **kwargs))
                ]
                results = await asyncio.gather(*processes)
                return results[-1]
            try:
                result = asyncio.run(run_and_check_all_async())
                self._print_result(result)
                return result
            except Exception as exception:
                self_cp.log_critical(f'{exception}')
                if raise_error:
                    raise
            finally:
                self_cp.play_bell()
        return main_loop

    def _get_main_loop_variables(
        self, env_prefix: str, args: Iterable[Any], kwargs: Mapping[str, Any]
    ) -> Tuple[TTask, Iterable[str], Mapping[str, Any]]:
        self_cp = copy.deepcopy(self)
        kwargs = {
            self._get_normalized_input_key(key): value
            for key, value in kwargs.items()
        }
        # ensure args and kwargs[_args] has the same value
        if len(args) == 0 and '_args' in kwargs:
            args = kwargs['_args']
        kwargs['_args'] = args
        # ensure kwargs are complete
        for task_input in self.inputs:
            kwarg_key = to_variable_name(task_input.name)
            if kwarg_key not in kwargs:
                kwargs[kwarg_key] = task_input.default
        # inject kwargs[_task]
        self_cp._set_keyval(input_map=kwargs, env_prefix=env_prefix)
        return self_cp, args, kwargs

    def _print_result(self, result: Any):
        '''
        Print result to stdout so that it can be processed further.
        e.g.: echo $(zrb show principle solid) > solid-principle.txt

        You need to override this method
        if you want to show the result differently.
        '''
        if result is None:
            return
        print(result)

    async def _loop_check(self, show_celebration: bool = False) -> bool:
        while not await self._cached_check():
            self.log_debug('Task is not ready')
            await asyncio.sleep(self.checking_interval)
        self.end_timer()
        self.log_info('Task is ready')
        if show_celebration:
            get_advertisement(advertisements).show()
            self.show_celebration()
        return True

    async def _cached_check(self) -> bool:
        if self._is_checked:
            self.log_debug('Skip checking, because checking flag has been set')
            return True
        check_result = await self._check()
        if check_result:
            self._is_checked = True
            self.log_debug('Set checking flag to True')
        return check_result

    async def _check(self) -> bool:
        '''
        Check current task readiness.
        - If self.checkers is defined,
          this will return True once every self.checkers is completed
        - Otherwise, this will return check method's return value.
        '''
        if len(self.checkers) == 0:
            return await self.check()
        check_processes: Iterable[asyncio.Task] = []
        for checker_task in self.checkers:
            check_processes.append(asyncio.create_task(checker_task.run()))
        await asyncio.gather(*check_processes)
        return True

    async def _run_all(self, *args: Any, **kwargs: Any) -> Any:
        processes: Iterable[asyncio.Task] = []
        # Add upstream tasks to processes
        for upstream_task in self.upstreams:
            processes.append(asyncio.create_task(
                upstream_task._run_all(**kwargs)
            ))
        # Add current task to processes
        processes.append(self._cached_run(*args, **kwargs))
        # Wait everything to complete
        results = await asyncio.gather(*processes)
        return results[-1]

    async def _cached_run(self, *args: Any, **kwargs: Any) -> Any:
        if self._is_executed:
            self.log_debug('Skip running, because execution flag has been set')
            return
        self.log_debug('Set execution flag to True')
        self._is_executed = True
        self.log_debug('Start running')
        # get upstream checker
        upstream_check_processes: Iterable[asyncio.Task] = []
        for upstream_task in self.upstreams:
            upstream_check_processes.append(asyncio.create_task(
                upstream_task._loop_check()
            ))
        # wait all upstream checkers to complete
        await asyncio.gather(*upstream_check_processes)
        # start running task
        result: Any
        local_kwargs = dict(kwargs)
        local_kwargs['_task'] = self
        while self.should_attempt():
            try:
                self.log_debug(
                    f'Invoke run with args: {args} and kwargs: {local_kwargs}'
                )
                result = await self.run(*args, **local_kwargs)
                break
            except Exception:
                if self.is_last_attempt():
                    raise
                attempt = self.get_attempt()
                self.log_error(f'Encounter error on attempt {attempt}')
                self.increase_attempt()
                await asyncio.sleep(self.retry_interval)
        self.mark_as_done()
        return result

    def _set_keyval(self, input_map: Mapping[str, Any], env_prefix: str):
        self._set_local_keyval(input_map=input_map, env_prefix=env_prefix)
        for upstream_task in self.upstreams:
            upstream_task._set_keyval(
                input_map=input_map, env_prefix=env_prefix
            )
        local_env_map = self.get_env_map()
        for checker_task in self.checkers:
            checker_task._set_keyval(
                input_map=input_map, env_prefix=env_prefix
            )
            # Checker task should be able to read local env
            checker_task._inject_env_map(local_env_map)



from zrb.task.docker_compose_task import DockerComposeTask

import pathlib
import os


def test_docker_compose_task_simple():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        compose_file=os.path.join(resource_path, 'docker-compose-simple.yml')
    )
    main_loop = docker_compose_task.create_main_loop()
    result = main_loop()
    assert 'Hello, Compose!' in result.output


def test_docker_compose_task_with_cwd_and_compose_file():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        cwd=resource_path,
        compose_file='docker-compose-simple.yml',
    )
    main_loop = docker_compose_task.create_main_loop()
    result = main_loop()
    assert 'Hello, Compose!' in result.output


def test_docker_compose_task_with_cwd():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        cwd=resource_path,
    )
    main_loop = docker_compose_task.create_main_loop()
    result = main_loop()
    assert 'Hello, Compose!' in result.output


def test_docker_compose_task_invalid_compose_file():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    is_error: bool = False
    try:
        DockerComposeTask(
            name='simple',
            cwd=resource_path,
            compose_file='non-existing.yml',
        )
    except Exception:
        is_error = True
    assert is_error


def test_docker_compose_task_simple_no_default_env():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        compose_file=os.path.join(
            resource_path, 'docker-compose-simple-no-default-env.yml'
        )
    )
    main_loop = docker_compose_task.create_main_loop()
    # run with env set
    os.environ['_ZRB_TEST_SIMPLE_NO_DEFAULT_MESSAGE'] = 'Good night'
    result_default = main_loop()
    assert 'Good night' in result_default.output


def test_docker_compose_task_simple_map_env():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        compose_file=os.path.join(
            resource_path, 'docker-compose-simple-map-env.yml'
        )
    )
    main_loop = docker_compose_task.create_main_loop()
    # run with no env
    if '_ZRB_TEST_SIMPLE_MAP_MESSAGE' in os.environ:
        del os.environ['_ZRB_TEST_SIMPLE_MAP_MESSAGE']
    result_default = main_loop()
    assert 'Good morning' in result_default.output
    # run with env set to empty string
    os.environ['_ZRB_TEST_SIMPLE_MAP_MESSAGE'] = ''
    result_default = main_loop()
    assert 'Good morning' in result_default.output
    # run with env set
    os.environ['_ZRB_TEST_SIMPLE_MAP_MESSAGE'] = 'Good night'
    result_custom = main_loop()
    assert 'Good night' in result_custom.output


def test_docker_compose_task_simple_list_env():
    dir_path = pathlib.Path(__file__).parent.absolute()
    resource_path = os.path.join(dir_path, 'resource')
    docker_compose_task = DockerComposeTask(
        name='simple',
        compose_file=os.path.join(
            resource_path, 'docker-compose-simple-list-env.yml'
        )
    )
    main_loop = docker_compose_task.create_main_loop()
    # run with no env
    if '_ZRB_TEST_SIMPLE_LIST_MESSAGE' in os.environ:
        del os.environ['_ZRB_TEST_SIMPLE_LIST_MESSAGE']
    result_default = main_loop()
    assert 'Good morning' in result_default.output
    # run with env set to empty string
    os.environ['_ZRB_TEST_SIMPLE_LIST_MESSAGE'] = ''
    result_default = main_loop()
    assert 'Good morning' in result_default.output
    # run with env set
    os.environ['_ZRB_TEST_SIMPLE_LIST_MESSAGE'] = 'Good night'
    result_custom = main_loop()
    assert 'Good night' in result_custom.output



from zrb import (
    DockerComposeTask, HTTPChecker, Env, runner
)
from zrb.builtin._group import project_group
import os

current_dir = os.path.dirname(__file__)
resource_dir = os.path.abspath(os.path.join(
    current_dir, '..', 'src', 'kebab-app-name'
))

compose_envs = [
    Env(
        name='MESSAGE',
        os_name='ENV_PREFIX_MESSAGE',
        default='Salve Mane'
    ),
    Env(
        name='CONTAINER_PORT',
        os_name='ENV_PREFIX_CONTAINER_PORT',
        default='3000'
    ),
    Env(
        name='HOST_PORT',
        os_name='ENV_PREFIX_HOST_PORT',
        default='httpPort'
    ),
]

start_snake_app_name_container = DockerComposeTask(
    name='start-kebab-app-name-container',
    description='Start human readable app name container',
    group=project_group,
    cwd=resource_dir,
    compose_cmd='up',
    compose_env_prefix='ENV_PREFIX',
    envs=compose_envs,
    checkers=[
        HTTPChecker(port='{{env.HOST_PORT}}')
    ]
)
runner.register(start_snake_app_name_container)

stop_snake_app_name_container = DockerComposeTask(
    name='stop-kebab-app-name-container',
    description='Stop human readable app name container',
    group=project_group,
    cwd=resource_dir,
    compose_cmd='stop',
    compose_env_prefix='ENV_PREFIX',
    envs=compose_envs,
)
runner.register(stop_snake_app_name_container)

remove_snake_app_name_container = DockerComposeTask(
    name='remove-kebab-app-name-container',
    description='Rumove human readable app name container',
    group=project_group,
    cwd=resource_dir,
    compose_cmd='down',
    compose_env_prefix='ENV_PREFIX',
    envs=compose_envs,
)
runner.register(remove_snake_app_name_container)

build_snake_app_name_image = DockerComposeTask(
    name='build-kebab-app-name-image',
    description='Build human readable app name image',
    group=project_group,
    cwd=resource_dir,
    compose_cmd='build',
    compose_env_prefix='ENV_PREFIX',
    envs=compose_envs,
)
runner.register(build_snake_app_name_image)

push_snake_app_name_image = DockerComposeTask(
    name='push-kebab-app-name-image',
    description='Push human readable app name image',
    group=project_group,
    cwd=resource_dir,
    compose_cmd='push',
    compose_env_prefix='ENV_PREFIX',
    envs=compose_envs,
)
runner.register(push_snake_app_name_image)



from typing import Any
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task_input.str_input import StrInput
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common import (
    project_dir_input, app_name_input, http_port_input, validate_project_dir,
    get_automation_module_name, validate_automation_name, register_module,
    get_default_task_replacements, get_default_task_replacement_middlewares,
    new_app_scaffold_lock
)
from ..project_task.add import (
    add_project_automation, register_project_upstream
)
from ....helper import util

import os

# Common definitions

current_dir = os.path.dirname(__file__)


@python_task(
    name='task-validate-create',
    inputs=[project_dir_input, app_name_input],
)
def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get(project_dir, 'app_name')
    validate_automation_name(project_dir, app_name)


replacements = get_default_task_replacements()
replacements.update({
    'ENV_PREFIX': '{{ util.coalesce(input.env_prefix, "MY").upper() }}'
})
copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        http_port_input,
        StrInput(
            name='env-prefix', prompt='Env prefix', default='MY'
        ),
    ],
    upstreams=[validate],
    replacements=replacements,
    replacement_middlewares=get_default_task_replacement_middlewares(),
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    scaffold_locks=[new_app_scaffold_lock]
)


@python_task(
    name='add-project-task',
    inputs=[project_dir_input, app_name_input],
    upstreams=[copy_resource]
)
def add_project_task(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    add_project_automation(project_dir)


@python_task(
    name='register-start',
    inputs=[project_dir_input, app_name_input],
    upstreams=[add_project_task]
)
def register_start(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    snake_app_name = util.to_snake_case(app_name)
    file_name = os.path.join(
        project_dir, '_automate', f'{snake_app_name}.py'
    )
    register_project_upstream(
        project_dir=project_dir,
        project_automation_file='start_project.py',
        project_automation_task_name='start',
        upstream_task_file=file_name,
        upstream_task_var=f'start_{snake_app_name}'
    )


@python_task(
    name='register-start-container',
    inputs=[project_dir_input, app_name_input],
    upstreams=[add_project_task]
)
def register_start_container(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    snake_app_name = util.to_snake_case(app_name)
    file_name = os.path.join(
        project_dir, '_automate', f'{snake_app_name}_container.py'
    )
    register_project_upstream(
        project_dir=project_dir,
        project_automation_file='start_project_containers.py',
        project_automation_task_name='start-containers',
        upstream_task_file=file_name,
        upstream_task_var=f'start_{snake_app_name}_container'
    )


@python_task(
    name='register-stop-container',
    inputs=[project_dir_input, app_name_input],
    upstreams=[add_project_task]
)
def register_stop_container(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    snake_app_name = util.to_snake_case(app_name)
    file_name = os.path.join(
        project_dir, '_automate', f'{snake_app_name}_container.py'
    )
    register_project_upstream(
        project_dir=project_dir,
        project_automation_file='stop_project_containers.py',
        project_automation_task_name='stop-containers',
        upstream_task_file=file_name,
        upstream_task_var=f'stop_{snake_app_name}_container'
    )


@python_task(
    name='register-remove-container',
    inputs=[project_dir_input, app_name_input],
    upstreams=[add_project_task]
)
def register_remove_container(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    snake_app_name = util.to_snake_case(app_name)
    file_name = os.path.join(
        project_dir, '_automate', f'{snake_app_name}_container.py'
    )
    register_project_upstream(
        project_dir=project_dir,
        project_automation_file='remove_project_containers.py',
        project_automation_task_name='remove-containers',
        upstream_task_file=file_name,
        upstream_task_var=f'remove_{snake_app_name}_container'
    )


@python_task(
    name='register-push-image',
    inputs=[project_dir_input, app_name_input],
    upstreams=[add_project_task]
)
def register_push_image(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    snake_app_name = util.to_snake_case(app_name)
    file_name = os.path.join(
        project_dir, '_automate', f'{snake_app_name}_container.py'
    )
    register_project_upstream(
        project_dir=project_dir,
        project_automation_file='push_project_images.py',
        project_automation_task_name='push-images',
        upstream_task_file=file_name,
        upstream_task_var=f'push_{snake_app_name}_image'
    )


@python_task(
    name='register-build-image',
    inputs=[project_dir_input, app_name_input],
    upstreams=[add_project_task]
)
def register_build_image(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    snake_app_name = util.to_snake_case(app_name)
    file_name = os.path.join(
        project_dir, '_automate', f'{snake_app_name}_container.py'
    )
    register_project_upstream(
        project_dir=project_dir,
        project_automation_file='build_project_images.py',
        project_automation_task_name='build-images',
        upstream_task_file=file_name,
        upstream_task_var=f'build_{snake_app_name}_image'
    )


register_task = Task(
    name='register-tasks',
    upstreams=[
        register_start,
        register_start_container,
        register_stop_container,
        register_remove_container,
        register_build_image,
        register_push_image
    ]
)


@python_task(
    name='simple-python-app',
    group=project_add_group,
    inputs=[project_dir_input, app_name_input],
    upstreams=[register_task],
    runner=runner
)
def add_simple_python_app(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir')
    app_name = kwargs.get('app_name')
    task.print_out(f'Register app: {app_name}')
    register_module(
        project_dir, get_automation_module_name(app_name)
    )
    register_module(
        project_dir, get_automation_module_name(f'{app_name}_container')
    )




from typing import Any
from ._group import md5_group
from ..task.decorator import python_task
from ..task_input.str_input import StrInput
from ..runner import runner

import hashlib

# Common definitions


@python_task(
    name='hash',
    group=md5_group,
    inputs=[
        StrInput(
            name='text',
            shortcut='t',
            description='Text',
            default=''
        )
    ],
    description='Hash md5',
    retry=0,
    runner=runner
)
def hash_text(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    hashed_text = hashlib.md5(text.encode()).hexdigest()
    return hashed_text


@python_task(
    name='sum',
    group=md5_group,
    inputs=[
        StrInput(
            name='file',
            shortcut='f',
            description='File',
            default=''
        )
    ],
    description='Sum md5 file',
    retry=0,
    runner=runner
)
def sum_file(*args: str, **kwargs: Any):
    file_path: str = kwargs.get('file', '')
    with open(file_path, "rb") as f:
        contents = f.read()
        hashed_text = hashlib.md5(contents).hexdigest()
        return hashed_text



from zrb import (
    CmdTask, HTTPChecker, Env, runner
)
from zrb.builtin._group import project_group
import os

current_dir = os.path.dirname(__file__)
resource_dir = os.path.abspath(os.path.join(
    current_dir, '..', 'src', 'kebab-app-name'
))


start_snake_app_name = CmdTask(
    name='start-kebab-app-name',
    description='Start human readable app name',
    group=project_group,
    cwd=os.path.join(resource_dir, 'src'),
    envs=[
        Env(
            name='MESSAGE',
            os_name='ENV_PREFIX_MESSAGE',
            default='Salve Mane'
        ),
        Env(
            name='PORT',
            os_name='ENV_PREFIX_HOST_PORT',
            default='httpPort'
        )
    ],
    cmd=[
        'if [ ! -d venv ]',
        'then',
        '  python -m venv venv',
        'fi',
        'pip install -r requirements.txt',
        'python main.py'
    ],
    checkers=[
        HTTPChecker(port='{{env.PORT}}')
    ]
)
runner.register(start_snake_app_name)



from typing import Callable, Iterable, Mapping
from typeguard import typechecked
from ...task.base_task import BaseTask
from ..util import (
    to_camel_case, to_pascal_case, to_kebab_case, to_snake_case,
    to_human_readable
)

import os


Replacement = Mapping[str, str]
ReplacementMiddleware = Callable[
    [BaseTask, Replacement],
    Replacement
]


@typechecked
def add_camel_key(
    new_key: str, old_key: str
) -> ReplacementMiddleware:
    def _add_camel_case(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        replacement[new_key] = to_camel_case(replacement.get(old_key, ''))
        return replacement
    return _add_camel_case


@typechecked
def add_pascal_key(
    new_key: str, old_key: str
) -> ReplacementMiddleware:
    def _add_pascal_case(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        replacement[new_key] = to_pascal_case(replacement.get(old_key, ''))
        return replacement
    return _add_pascal_case


@typechecked
def add_kebab_key(
    new_key: str, old_key: str
) -> ReplacementMiddleware:
    def _add_kebab_case(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        replacement[new_key] = to_kebab_case(replacement.get(old_key, ''))
        return replacement
    return _add_kebab_case


@typechecked
def add_snake_key(
    new_key: str, old_key: str
) -> ReplacementMiddleware:
    def _add_snake_case(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        replacement[new_key] = to_snake_case(replacement.get(old_key, ''))
        return replacement
    return _add_snake_case


@typechecked
def add_human_readable_key(
    new_key: str, old_key: str
) -> ReplacementMiddleware:
    def _add_human_readable(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        replacement[new_key] = to_human_readable(replacement.get(old_key, ''))
        return replacement
    return _add_human_readable


@typechecked
def coalesce(
    key: str, alternative_keys: Iterable[str] = [], default_value: str = '',
) -> ReplacementMiddleware:
    def _coalesce(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        if replacement[key] is not None and replacement[key] != '':
            return replacement
        for alt in alternative_keys:
            if replacement[alt] is not None and replacement[alt] != '':
                replacement[key] = replacement[alt]
                return replacement
        replacement[key] = default_value
        return replacement
    return _coalesce


@typechecked
def add_base_name_key(
    new_key: str, old_key: str
) -> ReplacementMiddleware:
    def _add_base_name(
        task: BaseTask, replacement: Replacement
    ) -> Replacement:
        replacement[new_key] = os.path.basename(replacement.get(old_key, ''))
        return replacement
    return _add_base_name
